import logging
import threading
import numpy as np
import multiprocessing as mp

from .hyperband import HyperbandScheduler
from .reporter import MODistStatusReporter
from ..utils.default_arguments import filter_by_key
from ..utils.mo_hbo_utils import prepare_sign_vector, uniform_from_unit_simplex
from ..utils.mo_hbo_utils import retrieve_pareto_front


__all__ = ['MOHyperbandScheduler']

logger = logging.getLogger(__name__)

_ARGUMENT_KEYS = {
    'objectives', 'scalarization_options'
}


class MOHyperbandScheduler(HyperbandScheduler):
    r"""Implements different variants of asynchronous multi-objective Hyperband
    by extending the standard HyperbandScheduler using random scalarizations. 
    The implemented method is described in detail in:

        Schmucker, Donini, Perrone, Zafar, Archambeau (2020)
        Multi-Objective Multi-Fidelity Hyperparameter Optimization with 
        Application to Fairness
        https://meta-learn.github.io/2020/papers/24_paper.pdf

    This class supports the full range of parameters of the standard
    HyperbandScheduler. The 'objectives' replaces the 'reward_attr' parameter.
    Here we only explain the multi-objective specific parameters. For the full 
    list of available options refer to HyperbandScheduler.

    Parameters
    ----------
    objectives : dict
        Dictionary with the names of objectives of interest. The corresponding
        values are allowed to be either "MAX" or "MIN" and indicate if an 
        objective is to be maximized or minimized.
    scalarization_options : dict
        Contains arguments for the underlying scalarization algorithm. 
        Available algorithms are "random_weights" and "parego".

    Examples
    --------
    examples/mo_hpo/mo_hyperband.ipynb
    """

    def __init__(self, train_fn, **kwargs):
        self._scalarization_options = kwargs["scalarization_options"]
        self._objectives = kwargs["objectives"]
        self._sign_vector = prepare_sign_vector(self._objectives)
        assert "_SCALARIZATION" not in kwargs["objectives"], "_SCALARIZATION" \
            "is a protected value. Please refrain from using it."
        kwargs["reward_attr"] = "_SCALARIZATION"
        super().__init__(train_fn, **filter_by_key(kwargs, _ARGUMENT_KEYS))

    def add_job(self, task, **kwargs):
        """Adding a training task to the scheduler.

        Args:
            task (:class:`autogluon.scheduler.Task`): a new training task
        """
        n_weights = self._scalarization_options.get('num_weights', 100)
        weights = [uniform_from_unit_simplex(len(self._objectives)) for _ in range(n_weights)]
        weights = [w * self._sign_vector for w in weights]
        reporter = MODistStatusReporter(self._objectives, weights, 
                self._scalarization_options, remote=task.resources.node)
        kwargs["reporter"] = reporter
        super().add_job(task, **kwargs)

    def get_pareto_front(self):
        """Retrieves the pareto efficient points discovered during the search
        process. Raises an error if called before search was conducted.

        Returns
        ----------
        front: list
            A list containing the Pareto efficient points among all points
            that were found during the search process.
        """
        assert len(self.finished_tasks) > 0, "Can only extract front after \
            jobs have been completed."
        pareto_front = retrieve_pareto_front(self.training_history, self._objectives)
        return pareto_front



from typing import NamedTuple
import time
import logging

from ..datatypes.tuning_job_state import TuningJobState

logger = logging.getLogger(__name__)


class ProfilingData(NamedTuple):
    id: int
    tag: str
    duration: float
    num_labeled: int
    num_pending: int
    fit_hyperparams: bool

    def to_tuple(self):
        return (self.id,
                self.tag,
                self.duration,
                self.num_labeled,
                self.num_pending,
                self.fit_hyperparams)

    @staticmethod
    def field_names():
        return ('id',
                'tag',
                'duration',
                'num_labeled',
                'num_pending',
                'fit_hyperparams')


class GPMXNetSimpleProfiler(object):
    def __init__(self):
        self.records = list()
        self.block = None
        self.start_time = dict()
        self.id_counter = 0

    def set_state(self, state: TuningJobState,
                  fit_hyperparams: bool):
        assert not self.start_time, \
            "Timers for these tags still running:\n{}".format(
                self.start_time.keys())
        self.block = ProfilingData(
            id=self.id_counter, tag='', duration=0.,
            num_labeled=len(state.candidate_evaluations),
            num_pending=len(state.pending_evaluations),
            fit_hyperparams=fit_hyperparams
        )
        self.id_counter += 1

    def start(self, tag: str):
        assert self.block is not None
        assert tag not in self.start_time, \
            "Timer for '{}' already running".format(tag)
        self.start_time[tag] = time.process_time()

    def stop(self, tag: str):
        assert tag in self.start_time, \
            "Timer for '{}' does not exist".format(tag)
        duration = time.process_time() - self.start_time[tag]
        self.records.append(
            self.block._replace(duration=duration, tag=tag))
        del self.start_time[tag]

    def clear(self):
        remaining_tags = list(self.start_time.keys())
        if remaining_tags:
            logger.warning("Timers for these tags not stopped (will be removed):\n{}".format(
                remaining_tags))
        self.start_time = dict()



from typing import List, NamedTuple, Set, Tuple, Iterator, Optional
import logging
import numpy as np
import itertools

from .base_classes import NextCandidatesAlgorithm, CandidateGenerator, ScoringFunction, LocalOptimizer, PendingCandidateStateTransformer, SurrogateModel
from .bo_algorithm_components import LBFGSOptimizeAcquisition
from .common import generate_unique_candidates
from ..autogluon.debug_log import DebugLogPrinter
from ..autogluon.gp_profiling import GPMXNetSimpleProfiler
from ..datatypes.common import Candidate, candidate_for_print
from ..utils.duplicate_detector import DuplicateDetector

logger = logging.getLogger(__name__)


class BayesianOptimizationAlgorithm(NamedTuple, NextCandidatesAlgorithm):
    """
    Core logic of the Bayesian optimization algorithm
    :param initial_candidates_generator: generator of candidates
    :param initial_scoring_function: scoring function used to rank the initial
        candidates.
        Note: If a batch is selected in one go (num_requested_candidates > 1,
        greedy_batch_selection = False), this function should encourage
        diversity among its top scorers. In general, greedy batch selection
        is recommended.
    :param num_initial_candidates: how many initial candidates to generate, if
        possible
    :param local_optimizer: local optimizer which starts from score minimizer.
        If a batch is selected in one go (not greedily), then local
        optimizations are started from the top num_requested_candidates ranked
        candidates (after scoring)
    :param pending_candidate_state_transformer: Once a candidate is selected, it
        becomes pending, and the state is transformed by appending information.
        This is done by the transformer.
        This is object is needed only if next_candidates goes through > 1 outer
        iterations (i.e., if greedy_batch_selection is True and
        num_requested_candidates > 1. Otherwise, None can be passed here.
        Note: Model updates (by the state transformer) for batch candidates beyond
        the first do not involve fitting hyperparameters, so they are usually
        cheap.
    :param blacklisted_candidates: set of tuples, candidates that should not be
        returned, because they are already labeled, currently pending, or have
        failed
    :param num_requested_candidates: number of candidates to return
    :param greedy_batch_selection: If True and num_requested_candidates > 1, we
        generate, order, and locally optimize for each single candidate to be
        selected. Otherwise (False), this is done just once, and
        num_requested_candidates are extracted in one go.
        Note: If this is True, pending_candidate_state_transformer is needed.
    :param duplicate_detector: used to make sure no candidates equal to already
        evaluated ones is returned
    :param profiler: If given, this is used for profiling parts in the code
    :param sample_unique_candidates: If True, we check that initial candidates
        sampled at random are unique and disjoint from the blacklist. See below.
    :param debug_log: If a DebugLogPrinter is passed here, it is used to write
        log messages

    Filtering out configs in blacklisted_candidates:
    If Candidate = ConfigSpace.Configuration, it turns out to be very expensive
    to ensure uniqueness of initial candidates, which is why
    sample_unique_candidates = False by default. In this case, make sure to
    use DuplicateDetectorIdentical or DuplicateDetectorEpsilon for
    duplicate_detector, which makes sure blacklisted_candidates are filtered
    out at the end. If you use DuplicateDetectorNoDetection, it is possible
    that the config returned is in blacklisted_candidates.

    """

    initial_candidates_generator: CandidateGenerator
    initial_candidates_scorer: ScoringFunction
    num_initial_candidates: int
    local_optimizer: LocalOptimizer
    pending_candidate_state_transformer: Optional[PendingCandidateStateTransformer]
    blacklisted_candidates: Set[Candidate]
    num_requested_candidates: int
    greedy_batch_selection: bool
    duplicate_detector: DuplicateDetector
    profiler: GPMXNetSimpleProfiler = None
    sample_unique_candidates: bool = False
    debug_log: Optional[DebugLogPrinter] = None

    # Note: For greedy batch selection (num_outer_iterations > 1), the
    # underlying SurrrogateModel changes with each new pending candidate. The
    # model changes are managed by pending_candidate_state_transformer. The
    # model has to be passed to both initial_candidates_scorer and
    # local_optimizer.
    def next_candidates(self) -> List[Candidate]:
        if self.greedy_batch_selection:
            # Select batch greedily, one candidate at a time, updating the
            # model in between
            num_outer_iterations = self.num_requested_candidates
            num_inner_candidates = 1
        else:
            # Select batch in one go
            num_outer_iterations = 1
            num_inner_candidates = self.num_requested_candidates
        assert num_outer_iterations == 1 or self.pending_candidate_state_transformer, \
            "Need pending_candidate_state_transformer for greedy batch selection"
        candidates = []
        model = None  # SurrogateModel, if num_outer_iterations > 1
        for outer_iter in range(num_outer_iterations):
            inner_candidates = self._get_next_candidates(
                num_inner_candidates, model=model)
            candidates.extend(inner_candidates)
            if outer_iter < num_outer_iterations - 1:
                # This is not the last outer iteration
                self.blacklisted_candidates.update(inner_candidates)
                # State transformer is used to produce new model
                # Note: We suppress fit_hyperpars for models obtained during
                # batch selection
                for candidate in inner_candidates:
                    self.pending_candidate_state_transformer.append_candidate(
                        candidate)
                model = self.pending_candidate_state_transformer.model(
                    skip_optimization=True)
        return candidates

    def _get_next_candidates(self, num_candidates: int,
                             model: Optional[SurrogateModel]):
        # generate a random candidates among which to pick the ones to be
        # locally optimized
        logger.log(15, "BayesOpt Algorithm: Generating initial candidates.")
        if self.profiler is not None:
            self.profiler.start('nextcand_genrandom')
        if self.sample_unique_candidates:
            # This can be expensive, depending on what type Candidate is
            initial_candidates = generate_unique_candidates(
                self.initial_candidates_generator,
                self.num_initial_candidates, self.blacklisted_candidates)
        else:
            initial_candidates = \
                self.initial_candidates_generator.generate_candidates_en_bulk(
                    self.num_initial_candidates)
        if self.profiler is not None:
            self.profiler.stop('nextcand_genrandom')
            self.profiler.start('nextcand_scoring')
        logger.log(15, "BayesOpt Algorithm: Scoring (and reordering) candidates.")
        if self.debug_log is not None:
            candidates_and_scores = _order_candidates(
                initial_candidates, self.initial_candidates_scorer,
                model=model, with_scores=True)
            initial_candidates = [cand for score, cand in candidates_and_scores]
            config = initial_candidates[0]
            top_scores = np.array([x for x, _ in candidates_and_scores[:5]])
            self.debug_log.set_init_config(config, top_scores)
        else:
            initial_candidates = _order_candidates(
                initial_candidates, self.initial_candidates_scorer,
                model=model)
        if self.profiler is not None:
            self.profiler.stop('nextcand_scoring')
            self.profiler.start('nextcand_localsearch')
        candidates_with_optimization = _lazily_locally_optimize(
            initial_candidates, self.local_optimizer, model=model)
        logger.log(15, "BayesOpt Algorithm: Selecting final set of candidates.")
        if self.debug_log is not None and \
                isinstance(self.local_optimizer, LBFGSOptimizeAcquisition):
            # We would like to get num_evaluations from the first run (usually
            # the only one). This requires peeking at the first entry of the
            # iterator
            peek = candidates_with_optimization.__next__()
            self.debug_log.set_num_evaluations(
                self.local_optimizer.num_evaluations)
            candidates_with_optimization = itertools.chain(
                [peek], candidates_with_optimization)
        candidates = _pick_from_locally_optimized(
            candidates_with_optimization, self.blacklisted_candidates,
            num_candidates, self.duplicate_detector)
        if self.profiler is not None:
            self.profiler.stop('nextcand_localsearch')
        return candidates


def _order_candidates(
        candidates: List[Candidate],
        scoring_function: ScoringFunction,
        model: Optional[SurrogateModel],
        with_scores: bool = False) -> List[Candidate]:
    if len(candidates) == 0:
        return []
    # scored in batch as this can be more efficient
    scores = scoring_function.score(candidates, model=model)
    sorted_list = sorted(zip(scores, candidates), key=lambda x: x[0])
    if with_scores:
        return sorted_list
    else:
        return [cand for score, cand in sorted_list]


def _lazily_locally_optimize(
        candidates: List[Candidate],
        local_optimizer: LocalOptimizer,
        model: Optional[SurrogateModel]) -> Iterator[Tuple[Candidate, Candidate]]:
    """
    Due to local deduplication we do not know in advance how many candidates
    we have to locally optimize, hence this helper to create a lazy generator
    of locally optimized candidates
    """
    for cand in candidates:
        yield cand, local_optimizer.optimize(cand, model=model)


# Note: If duplicate_detector is at least DuplicateDetectorIdentical, it will
# filter out candidates in blacklisted_candidates here. Such can in principle
# arise if sample_unique_candidates == False.
# This does not work if duplicate_detector is DuplicateDetectorNoDetection.
def _pick_from_locally_optimized(
        candidates_with_optimization: Iterator[Tuple[Candidate, Candidate]],
        blacklisted_candidates: Set[Candidate],
        num_candidates: int,
        duplicate_detector: DuplicateDetector) -> List[Candidate]:
    updated_blacklist = set(blacklisted_candidates)  # make a copy
    result = []
    for original_candidate, optimized_candidate in candidates_with_optimization:
        insert_candidate = None
        optimized_is_duplicate = duplicate_detector.contains(
            updated_blacklist, optimized_candidate)
        if optimized_is_duplicate:
            # in the unlikely case that the optimized candidate ended at a
            # place that caused a duplicate we try to return the original instead
            original_also_duplicate = duplicate_detector.contains(
                updated_blacklist, original_candidate)
            if not original_also_duplicate:
                insert_candidate = original_candidate
        else:
            insert_candidate = optimized_candidate
        if insert_candidate is not None:
            result.append(insert_candidate)
            updated_blacklist.add(insert_candidate)
        if len(result) == num_candidates:
            break

    return result



import numpy as np
from abc import ABC, abstractmethod
from typing import Tuple, Optional, NamedTuple

from ..tuning_algorithms.base_classes import SurrogateModel, AcquisitionFunction
from ..utils.density import get_quantiles


class HeadWithGradient(NamedTuple):
    hvals: np.array
    dh_dmean: np.array
    dh_dstd: np.array


class MeanStdAcquisitionFunction(AcquisitionFunction, ABC):
    """
    Base class for standard acquisition functions which depend on predictive
    mean and stddev. Subclasses have to implement the head and its derivatives
    w.r.t. mean and std:

        f(x, model) = h(mean, std, model.current_best())

    NOTE that acquisition functions will always be *minimized*!

    """
    def __init__(self, model: SurrogateModel):
        super(MeanStdAcquisitionFunction, self).__init__(model)
        assert 'mean' in model.keys_predict()
        assert 'std' in model.keys_predict()

    def compute_acq(self, inputs: np.ndarray,
                    model: Optional[SurrogateModel] = None) -> np.ndarray:
        if model is None:
            model = self.model
        assert 'mean' in model.keys_predict()
        assert 'std' in model.keys_predict()

        if inputs.ndim == 1:
            inputs = inputs.reshape((1, -1))
        predictions_list = model.predict(inputs)
        if self._head_needs_current_best():
            current_best_list = model.current_best()
        else:
            current_best_list = [None] * len(predictions_list)
        fvals_list = []
        # Loop over MCMC samples (if any)
        for predictions, current_best in zip(predictions_list, current_best_list):
            means = predictions['mean']
            stds = predictions['std']
            num_fantasies = means.shape[1] if means.ndim == 2 else 1
            if num_fantasies > 1:
                stds = stds.reshape((-1, 1))
                if current_best is not None:
                    assert current_best.size == num_fantasies, \
                        "mean.shape[1] = {}, current_best.size = {} (must be the same)".format(
                            num_fantasies, current_best.size)
                    current_best = current_best.reshape((1, -1))
            else:
                means = means.reshape((-1,))
                stds = stds.reshape((-1,))
                if current_best is not None:
                    current_best = current_best.reshape((1,))
            fvals = self._compute_heads(means, stds, current_best)
            # Average over fantasies if there are any
            if num_fantasies > 1:
                fvals = np.mean(fvals, axis=1)
            fvals_list.append(fvals)

        return np.mean(fvals_list, axis=0)

    def compute_acq_with_gradient(
            self, input: np.ndarray,
            model: Optional[SurrogateModel] = None) -> \
            Tuple[float, np.ndarray]:
        if model is None:
            model = self.model
        assert 'mean' in model.keys_predict()
        assert 'std' in model.keys_predict()

        predictions_list = model.predict(input.reshape(1, -1))
        if self._head_needs_current_best():
            current_best_list = model.current_best()
        else:
            current_best_list = [None] * len(predictions_list)
        fvals = []
        head_gradients = []

        for predictions, current_best in zip(predictions_list, current_best_list):
            mean = predictions['mean'].reshape((-1,))
            num_fantasies = mean.size
            std = predictions['std'].reshape((1,))
            if current_best is not None:
                assert current_best.size == num_fantasies
                current_best = current_best.reshape((-1,))
            head_result = self._compute_head_and_gradient(mean, std, current_best)
            fvals.append(np.mean(head_result.hvals))
            # Compose head_gradient for backward_gradient call. This involves
            # reshaping. Also, we are averaging over fantasies (if
            # num_fantasies > 1), which is not done in
            # _compute_head_and_gradient (dh_dmean, dh_dstd have the same shape
            # as mean), so have to divide by num_fantasies
            head_gradients.append({
                'mean': head_result.dh_dmean.reshape(
                    predictions['mean'].shape) / num_fantasies,
                'std': np.array([np.mean(head_result.dh_dstd)]).reshape(
                    predictions['std'].shape)})

        # Gradients are computed by the model
        gradient_list = model.backward_gradient(input, head_gradients)
        # Average over MCMC samples (if any)
        fval = np.mean(fvals)
        gradient = np.mean(gradient_list, axis=0)
        return fval, gradient

    @abstractmethod
    def _head_needs_current_best(self) -> bool:
        """
        :return: Is the current_best argument in _compute_head needed?
        """
        pass

    @abstractmethod
    def _compute_heads(
            self, means: np.ndarray, stds: np.ndarray,
            current_best: Optional[np.ndarray]) -> np.ndarray:
        """
        If mean has >1 columns, both std and current_best are supposed to be
        broadcasted. The return value has the same shape as mean.

        :param means: Predictive means, shape (n, nf) or (n,)
        :param stds: Predictive stddevs, shape (n, 1) or (n,)
        :param current_best: Incumbent, shape (1, nf) or (1,)
        :return: h(means, stds, current_best), same shape as means
        """
        pass

    @abstractmethod
    def _compute_head_and_gradient(
            self, mean: np.ndarray, std: np.ndarray,
            current_best: Optional[np.ndarray]) -> HeadWithGradient:
        """
        Computes both head value and head gradients, for a single input.

        :param mean: Predictive mean, shape (nf,)
        :param std: Predictive stddev, shape (1,)
        :param current_best: Incumbent, shape (nf,)
        :return: All HeadWithGradient fields have same shape as mean
        """
        pass


class EIAcquisitionFunction(MeanStdAcquisitionFunction):
    """
    Minus expected improvement acquisition function
    (minus because the convention is to always minimize acquisition functions)

    """
    def __init__(self, model: SurrogateModel, jitter: float = 0.01):
        super().__init__(model)
        self.jitter = jitter

    def _head_needs_current_best(self) -> bool:
        return True

    def _compute_heads(
            self, means: np.ndarray, stds: np.ndarray,
            current_best: Optional[np.ndarray]) -> np.ndarray:
        assert current_best is not None

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, means, stds)
        return (-stds) * (u * Phi + phi)

    def _compute_head_and_gradient(
            self, mean: np.ndarray, std: np.ndarray,
            current_best: Optional[np.ndarray]) -> HeadWithGradient:
        assert current_best is not None

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, mean, std)
        f_acqu = std * (u * Phi + phi)
        return HeadWithGradient(
            hvals=-f_acqu,
            dh_dmean=Phi,
            dh_dstd=-phi)


class LCBAcquisitionFunction(MeanStdAcquisitionFunction):
    """
    Lower confidence bound (LCB) acquisition function:

        h(mean, std) = mean - kappa * std

    """
    def __init__(self, model: SurrogateModel, kappa: float):
        super().__init__(model)
        assert kappa > 0, 'kappa must be positive'
        self.kappa = kappa

    def _head_needs_current_best(self) -> bool:
        return False

    def _compute_heads(
            self, means: np.ndarray, stds: np.ndarray,
            current_best: Optional[np.ndarray]) -> np.ndarray:
        return means - stds * self.kappa

    def _compute_head_and_gradient(
            self, mean: np.ndarray, std: np.ndarray,
            current_best: Optional[np.ndarray]) -> HeadWithGradient:
        ones_like_mean = np.ones_like(mean)
        return HeadWithGradient(
            hvals=mean - std * self.kappa,
            dh_dmean=ones_like_mean,
            dh_dstd=(-self.kappa) * ones_like_mean)



# Could eventually remove this code: Is this needed in unit tests?

"""
Object definitions that are used for testing.
"""

from typing import Iterator, Tuple, Dict
import numpy as np

from ..datatypes.common import StateIdAndCandidate
from ..datatypes.hp_ranges import HyperparameterRanges_Impl, \
    HyperparameterRangeContinuous, HyperparameterRangeInteger, \
    HyperparameterRangeCategorical, HyperparameterRanges
from ..datatypes.scaling import LogScaling, LinearScaling
from ..datatypes.tuning_job_state import TuningJobState
from ..gpautograd.constants import MCMCConfig, OptimizationConfig
from ..gpautograd.gp_regression import GaussianProcessRegression
from ..gpautograd.gpr_mcmc import GPRegressionMCMC
from ..gpautograd.kernel import Matern52, KernelFunction
from ..gpautograd.warping import WarpedKernel, Warping
from ..tuning_algorithms.base_classes import CandidateGenerator
from ..tuning_algorithms.default_algorithm import dictionarize_objective


def build_kernel(state: TuningJobState,
                 do_warping: bool = False) -> KernelFunction:
    dims, warping_ranges = dimensionality_and_warping_ranges(state.hp_ranges)
    kernel = Matern52(dims, ARD=True)
    if do_warping:
        return WarpedKernel(
            kernel=kernel, warping=Warping(dims, warping_ranges))
    else:
        return kernel


def default_gpmodel(
        state: TuningJobState, random_seed: int,
        optimization_config: OptimizationConfig) -> GaussianProcessRegression:
    return GaussianProcessRegression(
        kernel=build_kernel(state),
        optimization_config=optimization_config,
        random_seed=random_seed
    )


def default_gpmodel_mcmc(
        state: TuningJobState, random_seed: int,
        mcmc_config: MCMCConfig) -> GPRegressionMCMC:
    return GPRegressionMCMC(
        build_kernel=lambda: build_kernel(state),
        mcmc_config=mcmc_config,
        random_seed=random_seed
    )


def dimensionality_and_warping_ranges(hp_ranges: HyperparameterRanges) -> \
        Tuple[int, Dict[int, Tuple[float, float]]]:
    dims = 0
    warping_ranges = dict()
    # NOTE: This explicit loop over hp_ranges will fail if
    # HyperparameterRanges.hp_ranges is not implemented! Needs to be fixed if
    # it becomes an issue, either by moving the functionality here into
    # HyperparameterRanges, or by converting hp_ranges to
    # HyperparameterRanges_Impl, which supports the hp_ranges property.
    for hp_range in hp_ranges.hp_ranges:
        if not isinstance(hp_range, HyperparameterRangeCategorical):
            if isinstance(hp_range, HyperparameterRangeInteger):
                lower = int(round(hp_range.lower_bound))
                upper = int(round(hp_range.upper_bound))
            else:
                assert isinstance(hp_range, HyperparameterRangeContinuous)
                lower = float(hp_range.lower_bound)
                upper = float(hp_range.upper_bound)
            lower_internal = hp_range.to_ndarray(lower).item()
            upper_internal = hp_range.to_ndarray(upper).item()
            if upper_internal > lower_internal:  # exclude cases where max equal to min
                warping_ranges[dims] = (lower_internal, upper_internal)
            else:
                assert upper_internal == lower_internal
        dims += hp_range.ndarray_size()
    return dims, warping_ranges


class RepeatedCandidateGenerator(CandidateGenerator):
    """Generates candidates from a fixed set. Used to test the deduplication logic."""
    def __init__(self, n_unique_candidates: int):
        self.all_unique_candidates = [
            (1.0*j, j, "value_" + str(j))
            for j in range(n_unique_candidates)
        ]

    def generate_candidates(self) -> Iterator[StateIdAndCandidate]:
        i = 0
        while True:
            i += 1
            yield self.all_unique_candidates[i % len(self.all_unique_candidates)]


# Example black box function, with adjustable location of global minimum.
# Potentially could catch issues with optimizer, e.g. if the optimizer
# ignoring somehow candidates on the edge of search space.
# A simple quadratic function is used.
class Quadratic3d:
    def __init__(self, local_minima, active_metric, metric_names):
        # local_minima: point where local_minima is located
        self.local_minima = np.array(local_minima).astype('float')
        self.local_minima[0] = np.log10(self.local_minima[0])
        self.active_metric = active_metric
        self.metric_names = metric_names

    @property
    def search_space(self):
        return HyperparameterRanges_Impl(
            HyperparameterRangeContinuous('x', 1.0, 100.0, scaling=LogScaling()),
            HyperparameterRangeInteger('y', 0, 2, scaling=LinearScaling()),
            HyperparameterRangeCategorical('z', ('0.0', '1.0', '2.0'))
        )

    @property
    def f_min(self):
        return 0.0

    def __call__(self, candidate):
        p = np.array([float(hp) for hp in candidate])
        p[0] = np.log10(p[0])
        return dictionarize_objective(np.sum((self.local_minima - p) ** 2))



import logging
import time
from collections import OrderedDict
from copy import deepcopy
from typing import Tuple

import numpy as np
from tqdm.auto import tqdm

from autogluon.core.searcher import BaseSearcher, searcher_factory
from autogluon.core.utils import EasyDict

logger = logging.getLogger(__name__)


class LocalReporter:
    """
    Reporter implementation for LocalSequentialScheduler
    """

    def __init__(self, trial, searcher_config, training_history: dict, config_history: dict):
        self.trial = trial
        self.training_history = training_history
        self.training_history[trial] = []
        self.searcher_config = EasyDict(deepcopy(searcher_config))
        self.config_history = config_history
        self.trial_started = time.time()
        self.last_reported_time = self.trial_started
        self.last_result = None

    def __call__(self, *args, **kwargs):
        result = deepcopy(kwargs)
        if 'done' not in result:
            result['trial'] = self.trial

            now = time.time()
            result['time_this_iter'] = now - self.last_reported_time
            result['time_since_start'] = now - self.trial_started
            self.last_reported_time = now

            self.training_history[self.trial].append(result)

            if self.trial not in self.config_history:
                self.config_history[self.trial] = self.searcher_config
                if 'util_args' in self.searcher_config:
                    self.searcher_config.pop('util_args')

            self.last_result = result

    def terminate(self):
        pass  # compatibility


class LocalSequentialScheduler(object):
    """ Simple scheduler which schedules all HPO jobs in sequence without any parallelism.
    The next trial scheduling will be decided based on the available time left withing `time_out` setting
    and average time required for a trial to complete multiplied by the fill_factor (0.95) by default to
    accommodate variance in runtimes per HPO run.

    Parameters
    ----------
    train_fn : callable
        A task launch function for training.
    resource : dict
        Computation resources. For example, `{'num_cpus':2, 'num_gpus':1}`
    searcher : str
        Searcher (get_config decisions). If str, this is passed to
        searcher_factory along with search_options.
    search_options : dict
        If searcher is str, these arguments are passed to searcher_factory.
    num_trials : int
        Maximum number of jobs run in experiment. One of `num_trials`,
        `time_out` must be given.
    time_out : float
        If given, jobs are started only until this time_out (wall clock time).
        One of `num_trials`, `time_out` must be given.
    reward_attr : str
        Name of reward (i.e., metric to maximize) attribute in data obtained
        from reporter
    time_attr : str
        Name of resource (or time) attribute in data obtained from reporter.
        This attribute is optional for FIFO scheduling, but becomes mandatory
        in multi-fidelity scheduling (e.g., Hyperband).
        Note: The type of resource must be int.
    """

    def __init__(self, train_fn, searcher='auto', **kwargs):
        self.train_fn = train_fn
        self.training_history = None
        self.config_history = None
        self._reward_attr = kwargs['reward_attr']
        self.time_attr = kwargs.get('time_attr', None)
        self.resource = kwargs['resource']
        self.max_reward = kwargs.get('max_reward', None)
        self.searcher: BaseSearcher = self.get_searcher_(searcher, train_fn, **kwargs)
        self.init_limits_(kwargs)
        self.metadata = {
            'search_space': train_fn.kwspaces,
            'search_strategy': self.searcher,
            'stop_criterion': {
                'time_limits': self.time_out,
                'max_reward': self.max_reward},
            'resources_per_trial': self.resource
        }

    def init_limits_(self, kwargs):
        if kwargs.get('num_trials', None) is None:
            assert kwargs.get('time_out', None) is not None, "Need stopping criterion: Either num_trials or time_out"
        self.num_trials = kwargs.get('num_trials', 9999)
        self.time_out = kwargs.get('time_out', None)
        if self.num_trials is None:
            assert self.time_out is not None, \
                "Need stopping criterion: Either num_trials or time_out"

    def get_searcher_(self, searcher, train_fn, **kwargs) -> BaseSearcher:
        if searcher is 'local_sequential_auto':
            searcher = 'auto'
        scheduler_opts = {}
        if searcher is 'auto':
            searcher = 'bayesopt'
            scheduler_opts = {'scheduler': 'fifo'}
        search_options = kwargs.get('search_options', None)
        if isinstance(searcher, str):
            if search_options is None:
                search_options = dict()
            _search_options = search_options.copy()
            _search_options['configspace'] = train_fn.cs
            _search_options['reward_attribute'] = kwargs['reward_attr']
            _search_options['resource_attribute'] = kwargs.get('time_attr', None)
            # Adjoin scheduler info to search_options, if not already done by
            # subclass
            if 'scheduler' not in _search_options:
                _search_options['scheduler'] = 'fifo'
            searcher = searcher_factory(searcher, **{**scheduler_opts, **_search_options})
        else:
            assert isinstance(searcher, BaseSearcher)
        return searcher

    def run(self, **kwargs):
        """Run multiple trials given specific time and trial numbers limits.
        """
        self.searcher.configure_scheduler(self)

        self.training_history = OrderedDict()
        self.config_history = OrderedDict()

        trial_run_times = []
        time_start = time.time()

        r = range(self.num_trials)
        for i in (tqdm(r) if self.num_trials < 1000 else r):
            trial_start_time = time.time()
            is_failed = self.run_trial(task_id=i)
            trial_end_time = time.time()
            trial_run_times.append(np.NaN if is_failed else (trial_end_time - trial_start_time))

            if self.max_reward and self.get_best_reward() >= self.max_reward:
                logger.log(20, f'\tMax reward is reached')
                break

            if self.time_out is not None:
                avg_trial_run_time = np.nanmean(trial_run_times)
                avg_trial_run_time = 0 if np.isnan(avg_trial_run_time) else avg_trial_run_time
                if not self.has_enough_time_for_trial_(self.time_out, time_start, trial_start_time, trial_end_time, avg_trial_run_time):
                    logger.log(20, f'\tTime limit exceeded')
                    break

    @classmethod
    def has_enough_time_for_trial_(cls, time_out, time_start, trial_start_time, trial_end_time, avg_trial_run_time, fill_factor=0.95):
        """
        Checks if the remaining time is enough to run another trial.

        Parameters
        ----------
        time_out total
            timeout in m
        time_start
            trials start time
        trial_start_time
            last trial start time
        trial_end_time
            last trial end time
        avg_trial_run_time
            running average of all trial runs
        fill_factor: float
            discount of `avg_trial_run_time` allowed for a next trial. Default is 0.95 of `avg_trial_run_time`

        Returns
        -------
            True if there is enough time to run another trial give runs statistics and remaining time

        """
        time_spent = trial_end_time - time_start
        is_timeout_exceeded = time_spent >= time_out
        time_left = time_start + time_out - trial_end_time
        is_enough_time_for_another_trial = True
        if avg_trial_run_time:
            is_enough_time_for_another_trial = time_left > avg_trial_run_time * fill_factor
        return is_enough_time_for_another_trial and not is_timeout_exceeded

    @classmethod
    def get_average_trial_time_(cls, i, avg_trial_run_time, trial_start_time, time_end):
        trial_time = time_end - trial_start_time
        if avg_trial_run_time is None:
            avg_trial_run_time = trial_time
        else:
            avg_trial_run_time = ((avg_trial_run_time * i) + trial_time) / (i + 1)
        return avg_trial_run_time

    def run_trial(self, task_id=0) -> Tuple[bool, float, float]:
        """
        Start a trial with a given task_id

        Parameters
        ----------
        task_id
            task

        Returns
        -------
        is_failed: bool
            True if task completed successfully
        trial_start_time
            Trial start time
        trial_end_time
            Trial end time

        """
        searcher_config = self.searcher.get_config()
        reporter = LocalReporter(task_id, searcher_config, self.training_history, self.config_history)
        args = deepcopy(EasyDict(self.train_fn.kwvars))
        args['task_id'] = task_id
        self.searcher.register_pending(searcher_config)
        is_failed = False
        try:
            self.train_fn(args, config=searcher_config, reporter=reporter)
            if reporter.last_result:
                self.searcher.update(config=searcher_config, **reporter.last_result)
        except Exception as e:
            logger.error(f'Exception during a trial: {e}')
            self.searcher.evaluation_failed(config=searcher_config)
            reporter(traceback=e)
            is_failed = True
        return is_failed

    def join_jobs(self, timeout=None):
        pass  # Compatibility

    def get_best_config(self):
        """Get the best configuration from the finished jobs.
        """
        return self.searcher.get_best_config()

    def get_best_reward(self):
        """Get the best reward from the finished jobs.
        """
        return self.searcher.get_best_reward()

    def get_training_curves(self, filename=None, plot=False, use_legend=True):
        """Get Training Curves
        """
        if filename is None and not plot:
            logger.warning('Please either provide filename or allow plot in get_training_curves')
        import matplotlib.pyplot as plt

        eval_metric = self.__get_training_history_metric('eval_metric', default='validation_performance')
        sign_mult = int(self.__get_training_history_metric('greater_is_better', default=True)) * 2 - 1

        plt.ylabel(eval_metric)
        plt.xlabel(self.time_attr)
        plt.title("Performance vs Training-Time in each HPO Trial")
        for task_id, task_res in self.training_history.items():
            rewards = [x[self._reward_attr] * sign_mult for x in task_res]
            x = [x[self.time_attr] for x in task_res]
            plt.plot(x, rewards, label=f'task {task_id}')
        if use_legend:
            plt.legend(loc='best')
        if filename:
            logger.info(f'Saving Training Curve in {filename}')
            plt.savefig(filename)
        if plot:
            plt.show()

    def __get_training_history_metric(self, metric, default=None):
        for _, task_res in self.training_history.items():
            if task_res and metric in task_res[0]:
                return task_res[0][metric]
        return default

    def get_best_task_id(self):
        """Get the task id that results in the best configuration/best reward.

        If there are duplicated configurations, we return the id of the first one.
        """
        best_config = self.get_best_config()
        for task_id, config in self.config_history.items():
            if best_config == config:
                return task_id
        raise RuntimeError('The best config {} is not found in config history = {}. '
                           'This should never happen!'.format(best_config, self.config_history))



from abc import ABC, abstractmethod
from typing import List, Iterator, Iterable, Tuple, Type, Optional, Set, Dict
import numpy as np

from ..datatypes.common import Candidate
from ..datatypes.tuning_job_state import TuningJobState


class NextCandidatesAlgorithm:
    def next_candidates(self) -> List[Candidate]:
        # Not using ABC otherwise it will be difficult to create a subclass that also is a
        # NamedTuple(as both define their own metaclass)
        raise NotImplemented("Abstract method")


class CandidateGenerator(ABC):
    """
    Class to generate candidates from which to start the local minimization, typically random candidate
    or some form of more uniformly spaced variation, such as latin hypercube or sobol sequence
    """
    @abstractmethod
    def generate_candidates(self) -> Iterator[Candidate]:
        pass

    def generate_candidates_en_bulk(self, num_cands: int) -> List[Candidate]:
        raise NotImplementedError()


class SurrogateModel(ABC):
    def __init__(self, state: TuningJobState, active_metric: str,
                 random_seed: int):
        self.state = state
        self.random_seed = random_seed
        self.active_metric = active_metric

    @staticmethod
    def keys_predict() -> Set[str]:
        """
        Keys of signals returned by 'predict'.
        Note: In order to work with 'AcquisitionFunction' implementations,
        the following signals are required:

        - 'mean': Predictive mean
        - 'std': Predictive standard deviation

        :return:
        """
        return {'mean', 'std'}

    @abstractmethod
    def predict(self, inputs: np.ndarray) -> List[Dict[str, np.ndarray]]:
        """
        Given (n, d) matrix of encoded input points, returns signals which are
        statistics of the predictive distribution at these inputs. By default,
        signals are:

        - 'mean': Predictive means. If the model supports fantasizing with a
            number nf of fantasies, this has shape (n, nf), otherwise (n,)
        - 'std': Predictive stddevs, shape (n,)

        If the hyperparameters of the surrogate model are being optimized (e.g.,
        by empirical Bayes), the returned list has length 1. If its
        hyperparameters are averaged over by MCMC, the returned list has one
        entry per MCMC sample.
        """
        pass

    def predict_candidates(self, candidates: Iterable[Candidate]) -> \
            List[Dict[str, np.ndarray]]:
        """
        Convenience variant of 'predict', where the input is a list of n
        candidates, which are encoded to input points here.
        """
        return self.predict(self.state.hp_ranges.to_ndarray_matrix(candidates))

    @abstractmethod
    def current_best(self) -> List[np.ndarray]:
        """
        Returns the so-called incumbent, to be used in acquisition functions
        such as expected improvement. This is the minimum of predictive means
        (signal with key 'mean') at all current candidate locations (both
        state.candidate_evaluations and state.pending_evaluations).
        Normally, a scalar is returned, but if the model supports fantasizing
        and the state contains pending evaluations, there is one incumbent
        per fantasy sample, so a vector is returned.

        If the hyperparameters of the surrogate model are being optimized (e.g.,
        by empirical Bayes), the returned list has length 1. If its
        hyperparameters are averaged over by MCMC, the returned list has one
        entry per MCMC sample.

        :return: Incumbent
        """
        pass

    @abstractmethod
    def backward_gradient(
            self, input: np.ndarray,
            head_gradients: List[Dict[str, np.ndarray]]) -> List[np.ndarray]:
        """
        Computes the gradient nabla_x f of an acquisition function f(x),
        where x is a single input point. This is using reverse mode
        differentiation, the head gradients are passed by the acquisition
        function.

        If p = p(x) denotes the output of 'predict' for a single input point,
        'head_gradients' contains the head gradients nabla_p f. Its shape is
        that of p (where n=1).

        Lists have >1 entry if MCMC is used, otherwise they are all size 1.

        :param input: Single input point x, shape (d,)
        :param head_gradients: See above
        :return: Gradient nabla_x f (several if MCMC is used)
        """
        pass


class ScoringFunction(ABC):
    """
    Class to score candidates, typically combine an acquisition function with
    potentially Thompson sampling

    NOTE: it will be minimized, i.e. lower is better
    """
    @abstractmethod
    def score(self, candidates: Iterable[Candidate],
              model: Optional[SurrogateModel] = None) -> List[float]:
        """
        Requires multiple candidates, is this can be much quicker: we can use matrix operations

        lower is better
        """
        pass


class AcquisitionFunction(ScoringFunction):
    @abstractmethod
    def __init__(self, model: SurrogateModel):
        self.model = model

    @abstractmethod
    def compute_acq(self, inputs: np.ndarray,
                    model: Optional[SurrogateModel] = None) -> np.ndarray:
        """
        Note: If inputs has shape (d,), it is taken to be (1, d)

        :param inputs: Encoded input points, shape (n, d)
        :param model: If given, overrides self.model
        :return: Acquisition function values, shape (n,)
        """
        pass

    @abstractmethod
    def compute_acq_with_gradient(
            self, input: np.ndarray,
            model: Optional[SurrogateModel] = None) -> \
            Tuple[float, np.ndarray]:
        """
        For a single input point x, compute acquisition function value f(x)
        and gradient nabla_x f.

        :param input: Single input point x, shape (d,)
        :param model: If given, overrides self.model
        :return: f(x), nabla_x f
        """
        pass

    def score(self, candidates: Iterable[Candidate],
              model: Optional[SurrogateModel] = None) -> List[float]:
        if model is None:
            model = self.model
        inputs = model.state.hp_ranges.to_ndarray_matrix(candidates)
        return list(self.compute_acq(inputs, model=model))


class LocalOptimizer(ABC):
    """
    Class that tries to find a local candidate with a better score, typically using a local
    optimization method such as lbfgs. It would normally encapsulate an acquisition function and model

    """
    def __init__(self, state: TuningJobState, model: SurrogateModel,
                 acquisition_function_class: Type[AcquisitionFunction]):
        self.state = state
        self.model = model
        self.acquisition_function_class = acquisition_function_class

    @abstractmethod
    def optimize(self, candidate: Candidate,
                 model: Optional[SurrogateModel] = None) -> Candidate:
        """
        Run local optimization, starting from candidate.
        If model is given, it overrides self.model.

        :param candidate: Starting point
        :param model: See above
        :return: Candidate found by local optimization
        """
        pass


class PendingCandidateStateTransformer(ABC):
    """
    This concept is needed if HPO can deal with pending candidates, which
    remain unlabeled during further decision-making (e.g., batch decisions
    or asynchronous HPO). In this case, information for the pending candidate
    has to be added to the state.

    """
    @abstractmethod
    def append_candidate(self, candidate: Candidate):
        """
        Determines PendingEvaluation information for candidate and append to
        state.pending_evaluations.

        :param candidate: Novel pending candidate
        """
        pass

    @property
    @abstractmethod
    def state(self) -> TuningJobState:
        """
        :return: Current TuningJobState
        """
        pass

    @abstractmethod
    def model(self, **kwargs) -> SurrogateModel:
        """
        In general, the model is computed here on demand, based on the current
        state, unless the state has not changed since the last 'model' call.

        :return: Surrogate model for current state
        """
        pass
